
// Discrete.fs - Aivika

// Copyright (c) 2009, 2010, 2011  David Sorokin. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials
//     provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
// THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.

//
// Author: David Sorokin <david.sorokin@gmail.com>
//

namespace Maritegra.Aivika

open System
open System.Collections.Generic

open SD

type Env () =

    let dict = ref Map.empty
    let dictLock = new obj ()
    
    let add t c (q: PriorityQueue<_>) =
        q.Enqueue (t, c)

    let call r s rt te (h: PriorityQueue<_>) =
        let mutable f = true
        while f && (not h.IsEmpty) do
            let (t2, c2: Dynamics<_>) = h.Front
            match !rt with
                | Some t when t2 < t ->
                    failwithf "The time value is too small" 
                | _ -> ()
            if t2 <= te then
                rt := Some t2
                h.Dequeue ()
                let n2  = int (floor ((t2 - s.StartTime) / s.DT))
                let ph2 = -1
                let ps2 = { Run = r; Specs = s; Time = t2; Iteration = n2; Phase = ph2 }
                (c2 <|- ps2) ()
            else
                f <- false
            
    let run =
        Dynamics (fun ps ->
            let r = ps.Run
            let s = ps.Specs
            let t = ps.Time
            let k = r.Id
            if Map.containsKey k !dict then
                let (f, rt, l) = Map.find k !dict
                if !f = 0 then
                    incr f
                    call r s rt t l
                    decr f)

    member x.Run = run
    
    member x.DispatchD (t: Dynamics<float>, c: Dynamics<unit -> unit>) =
        let u = t in Dynamics (fun ps ->
            let r  = ps.Run
            let k  = r.Id
            let u' = u <|- ps
            if Map.containsKey k !dict then
                let (f, m, q) = Map.find k !dict
                add u' c q
            else
                let q = new PriorityQueue<_> ()
                add u' c q
                // -- dict --
                let t0 = starttime <|- ps
                lock dictLock (fun () -> 
                    dict := Map.add k (ref 0, ref (Some t0), q) !dict)
                r.Disposed.Add (fun () ->
                    lock dictLock (fun () -> 
                        dict := Map.remove k !dict)))
            
    member x.Dispatch (t, c) = x.DispatchD (t, Dynamics.delay c)

and [<AbstractClass>] Process (e: Env) =

    let dict = ref Map.empty
    let dictLock = new obj ()
    let dictFin = ref Map.empty
    
    let finalize (r: Run) =
        let k = r.Id
        if not (Map.containsKey k !dictFin) then
            lock dictLock (fun () ->
                dictFin := Map.add k () !dictFin)
            r.Disposed.Add (fun () ->
                lock dictLock (fun () ->
                    dict := Map.remove k !dict
                    dictFin := Map.remove k !dictFin))
    
    let peek (r: Run) =
        let k = r.Id
        if Map.containsKey k !dict then
            let w = Map.find k !dict in !w
        else
            None
    
    let pop (r: Run) =
        let k = r.Id
        if Map.containsKey k !dict then
            let w = Map.find k !dict 
            match !w with
                | None -> 
                    None
                | Some c ->
                    let v = !w
                    w := None
                    v
        else
            None
            
    let push (r: Run) c =
        let k = r.Id
        if Map.containsKey k !dict then
            let w = Map.find k !dict
            match !w with
                | None -> w := Some c
                | Some c -> failwithf "Cannot passivate twice the same process"
        else
            let w = ref (Some c)
            lock dictLock (fun () ->
                dict := Map.add k w !dict)
            finalize r
            
    let isPassivated = 
        DynamicsCont (fun c ->
            Dynamics (fun ps ->
                let r = ps.Run
                let a = Option.isSome (peek r)
                let (cont, econt) = c <|- ps
                in cont a))
    
    let passivate =
        DynamicsCont (fun c ->
            Dynamics (fun ps ->
                let r = ps.Run
                push r (Dynamics.unzipFst c)))
    
    let reactivate =
        DynamicsCont (fun c -> 
            Dynamics (fun ps -> 
                let r = ps.Run
                match pop r with
                    | Some c2 ->
                        let m = e.DispatchD (time, Dynamics.unzipFst c)
                        in m <|- ps      // N.B. think over whether it should be protected!
                        let cont2 = c2 <|- ps
                        in cont2 ()
                    | None ->
                        let (cont, econt) = c <|- ps
                        in cont ()))
            
    member x.Env = e
                
    member x.Start (t: float) =
        let m = dynamicscont {
            do! DynamicsCont (fun c -> e.DispatchD (eta t, Dynamics.unzipFst c))
            do! x.Activate   // it will be called later by value e
        } 
        Dynamics.once <| DynamicsCont.run m (fun () -> ())
                
    member x.StartD (t) =
        let m = dynamicscont {
            do! DynamicsCont (fun c -> e.DispatchD (t, Dynamics.unzipFst c))
            do! x.Activate   // it will be called later by value e
        } 
        Dynamics.once <| DynamicsCont.run m (fun () -> ())
    
    member x.Hold (t: float) = 
        DynamicsCont (fun c -> e.DispatchD (t + time, Dynamics.unzipFst c))
    
    member x.HoldD (t) = 
        DynamicsCont (fun c -> e.DispatchD (t + time, Dynamics.unzipFst c))
    
    abstract Activate: DynamicsCont<unit>
        
    member x.IsPassivated = isPassivated
    member x.Passivate = passivate
    member x.Reactivate = reactivate

and Resource (e: Env, number) =

    let dict = ref Map.empty
    let dictLock = new obj ()
    let dictQueues = ref Map.empty
    let dictFin = ref Map.empty
    
    let finalize (r: Run) =
        let k = r.Id
        if not (Map.containsKey k !dictFin) then
            lock dictLock (fun () ->
                dictFin := Map.add k () !dictFin)
            r.Disposed.Add (fun () ->
                lock dictLock (fun () ->
                    dict := Map.remove k !dict
                    dictQueues := Map.remove k !dictQueues
                    dictFin := Map.remove k !dictFin))

    let num =
        DynamicsCont (fun c ->
            Dynamics (fun ps ->
                let r = ps.Run
                let k = r.Id
                if Map.containsKey k !dict then
                    let ri = Map.find k !dict
                    let (cont, econt) = c <|- ps
                    in cont !ri
                else
                    let (cont, econt) = c <|- ps
                    in cont number))
    
    let request =
        DynamicsCont (fun c ->
            Dynamics (fun ps ->
                let r = ps.Run
                let z = ref None
                let k = r.Id
                if Map.containsKey k !dict then
                    let ri = Map.find k !dict
                    if !ri > 0 then
                        decr ri
                        z := Some (Dynamics.unzipFst c)
                elif number > 0 then
                    lock dictLock (fun () ->
                        dict := Map.add k (ref (number - 1)) !dict)
                    finalize r
                    z := Some (Dynamics.unzipFst c)
                match !z with
                    | None -> 
                        if Map.containsKey k !dictQueues then
                            let l = (Map.find k !dictQueues) : LinkedList<_>
                            l.AddLast (Dynamics.unzipFst c) |> ignore
                        else
                            let l = new LinkedList<_> ()
                            lock dictLock (fun () ->
                                dictQueues := Map.add k l !dictQueues)
                            finalize r
                            l.AddLast (Dynamics.unzipFst c) |> ignore
                    | Some c2 -> 
                        (c2 <|- ps) ()))
    
    let release =
        DynamicsCont (fun c ->
            Dynamics (fun ps ->
                let r = ps.Run
                let z = ref None
                let k = r.Id
                if Map.containsKey k !dict then
                    let ri = Map.find k !dict
                    assert ((0 <= !ri) && (!ri < number))
                    incr ri
                    if Map.containsKey k !dictQueues then
                        let l = Map.find k !dictQueues
                        if l.Count > 0 then
                            let c2 = l.First.Value
                            l.RemoveFirst ()
                            assert (!ri > 0)
                            decr ri
                            z := Some c2
                else
                    failwithf "The variable was not requested"
                match !z with
                    | Some c2 -> 
                        let m = e.DispatchD (time, c2)
                        m <|- ps     // N.B. think over whether it should be protected!
                    | None -> ()
                let (cont, econt) = c <|- ps
                cont ()))
    
    member x.Env = e
    
    member x.N = num
    
    member x.Request = request
    member x.Release = release
                
and Var<'a> (e: Env, i: Dynamics<'a>) =

    let dict = ref Map.empty
    let dictLock = new obj ()
    let dictFin = ref Map.empty
    
    let finalize (r: Run) =
        let k = r.Id
        if not (Map.containsKey k !dictFin) then
            lock dictLock (fun () ->
                dictFin := Map.add k () !dictFin)
            r.Disposed.Add (fun () ->
                lock dictLock (fun () ->
                    dict := Map.remove k !dict
                    dictFin := Map.remove k !dictFin))

    let me = e.Run
    
    let find (ps: Parameters) =
        let r = ps.Run
        let s = ps.Specs
        let k = r.Id
        if Map.containsKey k !dict then
            Map.find k !dict
        else
            let t = s.StartTime
            let a = i <|- { ps with Time = t; Iteration = 0; Phase = 0 }
            let x = ResizeArray<_> ()
            let y = ResizeArray<_> ()
            let z  = (x, y)
            x.Add (t)
            y.Add (a)
            lock dictLock (fun () ->
                dict := Map.add k z !dict)
            finalize r
            z
        
    let value = 
        Dynamics (fun ps ->
            let t = ps.Time
            me <|- ps
            let (x: ResizeArray<_>, y: ResizeArray<_>) = find ps
            let i = x.Count - 1
            if x.[i] <= t then
                y.[i]
            else
                let i = x.BinarySearch (t)
                if i >= 0 then
                    y.[i]
                else
                    let i = (~~~i) - 1
                    assert (i >= 0)
                    y.[i])
                
    new (e: Env, i: 'a) = Var<'a> (e, eta i)
                
    member x.Env = e
    
    member x.Value = value
    
    member x.SetValue (a: 'a) =
        Dynamics (fun ps ->
            let t = ps.Time
            me <|- ps
            let (x, y) = find ps
            let i = x.Count - 1
            if t < x.[i] then
                failwithf "Cannot update the past data"
            elif t = x.[i] then
                y.[i] <- a
            else
                x.Add (t)
                y.Add (a))

and MemorylessVar<'a> (e: Env, i: Dynamics<'a>) =

    let dict = ref Map.empty
    let dictLock = new obj ()
    let dictFin = ref Map.empty
    
    let finalize (r: Run) =
        let k = r.Id
        if not (Map.containsKey k !dictFin) then
            lock dictLock (fun () ->
                dictFin := Map.add k () !dictFin)
            r.Disposed.Add (fun () ->
                lock dictLock (fun () ->
                    dict := Map.remove k !dict
                    dictFin := Map.remove k !dictFin))

    let me = e.Run
    
    let find (ps: Parameters) =
        let r = ps.Run
        let s = ps.Specs
        let k = r.Id
        if Map.containsKey k !dict then
            Map.find k !dict
        else
            let t = s.StartTime
            let a = i <|- { ps with Time = t; Iteration = 0; Phase = 0 }
            let z = ref (t, a)
            lock dictLock (fun () ->
                dict := Map.add k z !dict)
            finalize r
            z
        
    let value = 
        Dynamics (fun ps ->
            let t = ps.Time
            me <|- ps
            let z = find ps
            let (t', a) = !z
            if t < t' then
                failwithf "Cannot request the past data"
            a)
                
    new (e: Env, i: 'a) = MemorylessVar<'a> (e, eta i)
                
    member x.Env = e
    
    member x.Value = value
    
    member x.SetValue (a: 'a) =
        Dynamics (fun ps ->
            let t = ps.Time
            me <|- ps
            let z = find ps
            let (t', _) = !z
            if t < t' then
                failwithf "Cannot update the past data"
            else
                z := (t, a))

[<CompilationRepresentation (CompilationRepresentationFlags.ModuleSuffix)>]
module Var =
    
    [<CompiledName ("Env")>]
    let env (v: Var<_>) = v.Env
    
    [<CompiledName ("Value")>]
    let value (v: Var<_>) = v.Value
    
    [<CompiledName ("SetValue")>]
    let setValue a (v: Var<_>) = v.SetValue (a)
    
    [<CompiledName ("IncValue")>]
    let inline incValue (a:'a when 'a: (static member (+): 'a * 'a -> 'a)) (v: Var<'a>) = dynamics {
        let! x = v.Value
        do! v.SetValue (x + a)
    }
    
    [<CompiledName ("DecValue")>]
    let inline decValue (a:'a when 'a: (static member (-): 'a * 'a -> 'a)) (v: Var<'a>) = dynamics {
        let! x = v.Value
        do! v.SetValue (x - a)
    }

[<CompilationRepresentation (CompilationRepresentationFlags.ModuleSuffix)>]
module MemorylessVar =
    
    [<CompiledName ("Env")>]
    let env (v: MemorylessVar<_>) = v.Env
    
    [<CompiledName ("Value")>]
    let value (v: MemorylessVar<_>) = v.Value
    
    [<CompiledName ("SetValue")>]
    let setValue a (v: MemorylessVar<_>) = v.SetValue (a)
    
    [<CompiledName ("IncValue")>]
    let inline incValue (a:'a when 'a: (static member (+): 'a * 'a -> 'a)) (v: MemorylessVar<'a>) = dynamics {
        let! x = v.Value
        do! v.SetValue (x + a)
    }
    
    [<CompiledName ("DecValue")>]
    let inline decValue (a:'a when 'a: (static member (-): 'a * 'a -> 'a)) (v: MemorylessVar<'a>) = dynamics {
        let! x = v.Value
        do! v.SetValue (x - a)
    }
