
// Dynamics.fs - Aivika

// Copyright (c) 2009, 2010, 2011  David Sorokin. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials
//     provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
// THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.

//
// Author: David Sorokin <david.sorokin@gmail.com>
//

namespace Maritegra.Aivika

open System
open System.Collections.Generic

type internal Time = float

type internal Iteration = int

type internal Phase = int

type Method = Euler | RungeKutta2 | RungeKutta4

type Randomness = SimpleRnd | StrongRnd

type Specs = 
    { StartTime: float; StopTime: float; DT: float; 
      Method: Method; Randomness: Randomness }

[<RequireQualifiedAccess>]
[<CompilationRepresentation (CompilationRepresentationFlags.ModuleSuffix)>]
module internal Specs =

    let iterations specs =
        1 + int (ceil ((specs.StopTime - specs.StartTime) / specs.DT)) : Iteration
        
    let phases specs = 
        match specs.Method with
            | Euler -> 1 : Phase
            | RungeKutta2 -> 2 : Phase
            | RungeKutta4 -> 4 : Phase

    let time specs n ph =
        let delta m ph = 
            match m, ph with
                | Euler, 0 -> 0.0
                | RungeKutta2, 0 -> 0.0
                | RungeKutta2, 1 -> specs.DT
                | RungeKutta4, 0 -> 0.0
                | RungeKutta4, 1 -> specs.DT / 2.0
                | RungeKutta4, 2 -> specs.DT / 2.0
                | RungeKutta4, 3 -> specs.DT
                | _  -> failwith "Incorrect method and/or phase"
        in specs.StartTime + float (n) * specs.DT + delta specs.Method ph
        
    let inline neighborhood specs (t: Time) t' =
        abs (t - t') <= specs.DT / 1.0e6

type Specs with
    member internal x.Iterations = Specs.iterations x
    member internal x.Phases = Specs.phases x

type internal Run () =

    static let mutable idgen = 0L
    static let idlock = new obj ()
    
    let id = lock idlock (fun () ->
        idgen <- idgen + 1L
        idgen)
    
    let mutable disposed = false
    let disposedEvent = new Event<_>()

    member x.Id = id
    member x.Disposed = disposedEvent.Publish
    member x.IsDisposed = disposed
    
    interface IDisposable with
        member x.Dispose() =
            if not disposed then
                disposedEvent.Trigger()
                disposed <- true

type internal Parameters =
    { Run: Run; Specs: Specs; Time: float;
      Iteration: Iteration; Phase: Phase }

type internal DynamicsFun<'a> = Parameters -> 'a

[<Sealed>] 
type Dynamics<'a> (f: DynamicsFun<'a>) =
    member internal d.Fun = f
    
[<AutoOpen>]
module internal DynamicsOp =
    let inline (<|-) (d: Dynamics<'a>) x = d.Fun x
    
module DynamicsBuilderImpl =

    let inline returnD a =
        Dynamics (fun ps -> a)

    let inline bindD m k =
        Dynamics (fun ps ->
            let a = m <|- ps
            let m' = k a
            m' <|- ps)

    //  letD a k = bindD (returnD a) k
    let inline letD a k =
        Dynamics (fun ps ->
            let m' = k a
            m' <|- ps)

    //  delayD k = bindD (returnD ()) k
    let inline delayD k =
        Dynamics (fun ps ->
            let m' = k ()
            m' <|- ps)
            
    //  combineD m1 m2 = bindD m1 (fun () -> m2)
    let inline combineD m1 m2 =
        Dynamics (fun ps ->
            m1 <|- ps
            m2 <|- ps)
            
    let zeroD = Dynamics (fun ps -> ())
    
    let inline forD es k  =
        Dynamics (fun ps ->
            for e in es do
                (k e) <|- ps)
                
    let inline whileD p m =
        Dynamics (fun ps ->
            while p () do
                m <|- ps)
                
    let inline usingD (a: 'a when 'a :> IDisposable) k =
        Dynamics (fun ps ->
            try
                (k a) <|- ps
            finally
                a.Dispose ())
                
    let inline tryFinallyD m f =
        Dynamics (fun ps ->
            try
                m <|- ps
            finally
                f ())
                
    let inline tryWithD m k =
        Dynamics (fun ps ->
            try
                m <|- ps
            with
            | e -> (k e) <|- ps)
            
open DynamicsBuilderImpl

[<Sealed>]
type DynamicsBuilder() =

    member d.Return (a)  = returnD a
    member d.ReturnFrom (m: Dynamics<_>) = m
    member d.Bind (m, k) = bindD m k
    member d.Delay (k)   = delayD k
    member d.Zero () = zeroD
    member d.Combine (m1, m2) = combineD m1 m2
    member d.For (es, k) = forD es k
    member d.While (p, m) = whileD p m
    member d.Using (a, k) = usingD a k
    member d.TryFinally (m, f) = tryFinallyD m f
    member d.TryWith (m, k) = tryWithD m k

[<AutoOpen>]
module DynamicsWorkflow =
    let dynamics = new DynamicsBuilder()

[<RequireQualifiedAccess>]
[<CompilationRepresentation (CompilationRepresentationFlags.ModuleSuffix)>]
module Dynamics =

    [<CompiledName ("Run")>]
    let run d specs = seq {
    
        use r = new Run ()
        for i = 0 to (Specs.iterations specs - 1) do
            let ps = { Run = r; Specs = specs; 
                       Time = (Specs.time specs i 0);
                       Iteration = i; Phase = 0 }
            yield d <|- ps
    }

    [<CompiledName ("RunLast")>]
    let runLast d specs =
    
        use r = new Run ()
        let i = Specs.iterations specs - 1
        let ps = { Run = r; Specs = specs; 
                   Time = (Specs.time specs i 0);
                   Iteration = i; Phase = 0 }
        in d <|- ps
        
    [<CompiledName ("RunWhile")>]
    let runWhile c d specs = seq {
    
        use r = new Run ()
        let f = ref true
        let i = ref 0
        let j = (Specs.iterations specs - 1)
        while !f && (!i <= j) do
            let ps = { Run = r; Specs = specs; 
                       Time = (Specs.time specs !i 0);
                       Iteration = !i; Phase = 0 }
            let c' = c <|- ps
            if not c' then 
                f := false
            else
                yield d <|- ps
                incr i
    }

    [<CompiledName ("Eta")>]
    let eta a = Dynamics (fun ps -> a)
    
    [<CompiledName ("Mu")>]
    let mu mm = Dynamics (fun ps ->
        let m = mm <|- ps
        in m <|- ps)
    
    [<CompiledName ("Map")>]
    let map f m = Dynamics (fun ps ->
        let a = m <|- ps
        in f a)
    
    [<CompiledName ("Ap")>]
    let ap mf m = Dynamics (fun ps ->
        let f = mf <|- ps
        let a = m <|- ps
        in f a)
    
    [<CompiledName ("Lift2")>]
    let lift2 f m1 m2 = Dynamics (fun ps ->
        let a1 = m1 <|- ps
        let a2 = m2 <|- ps
        in f a1 a2)
    
    [<CompiledName ("Lift3")>]
    let lift3 f m1 m2 m3 = Dynamics (fun ps ->
        let a1 = m1 <|- ps
        let a2 = m2 <|- ps
        let a3 = m3 <|- ps
        in f a1 a2 a3)
    
    [<CompiledName ("Lift4")>]
    let lift4 f m1 m2 m3 m4 = Dynamics (fun ps ->
        let a1 = m1 <|- ps
        let a2 = m2 <|- ps
        let a3 = m3 <|- ps
        let a4 = m4 <|- ps
        in f a1 a2 a3 a4)
    
    [<CompiledName ("Lift5")>]
    let lift5 f m1 m2 m3 m4 m5 = Dynamics (fun ps ->
        let a1 = m1 <|- ps
        let a2 = m2 <|- ps
        let a3 = m3 <|- ps
        let a4 = m4 <|- ps
        let a5 = m5 <|- ps
        in f a1 a2 a3 a4 a5)
        
    [<CompiledName ("Generate")>]
    let generate a mf = Dynamics (fun ps ->
        let f = mf <|- ps
        in f a)
        
    [<CompiledName ("Zip")>]
    let zip m1 m2 = Dynamics (fun ps ->
        let a1 = m1 <|- ps
        let a2 = m2 <|- ps
        in (a1, a2)) 
        
    [<CompiledName ("Zip3")>]
    let zip3 m1 m2 m3 = Dynamics (fun ps ->
        let a1 = m1 <|- ps
        let a2 = m2 <|- ps
        let a3 = m3 <|- ps
        in (a1, a2, a3)) 
        
    [<CompiledName ("Unzip")>]
    let unzip m = 
        let m1 = Dynamics (fun ps -> fst (m <|- ps))
        let m2 = Dynamics (fun ps -> snd (m <|- ps))
        in (m1, m2)
        
    [<CompiledName ("Unzip3")>]
    let unzip3 m =
        let fst (x1, x2, x3) = x1
        let snd (x1, x2, x3) = x2
        let thd (x1, x2, x3) = x3
        let m1 = Dynamics (fun ps-> fst (m <|- ps))
        let m2 = Dynamics (fun ps -> snd (m <|- ps))
        let m3 = Dynamics (fun ps -> thd (m <|- ps))
        in (m1, m2, m3)
    
    [<CompiledName ("UnzipFst")>]
    let unzipFst m = Dynamics (fun ps -> fst (m <|- ps))
    
    [<CompiledName ("UnzipSnd")>]
    let unzipSnd m = Dynamics (fun ps -> snd (m <|- ps))
        
    [<CompiledName ("Delay")>]
    let delay m = Dynamics (fun ps ->
        fun () -> m <|- ps)
    
    [<CompiledName ("OfList")>]
    let ofList (ms: Dynamics<_> list) = Dynamics (fun ps ->
        [ for m in ms -> m <|- ps ])
    
    [<CompiledName ("OfArray")>]
    let ofArray (ms: Dynamics<_> []) = Dynamics (fun ps ->
        [| for m in ms -> m <|- ps |])
                  
    [<CompiledName ("SeqList")>]
    let seqList (ms: #seq<Dynamics<_>>) = Dynamics (fun ps ->
        [ for m in ms -> m <|- ps ])
                  
    [<CompiledName ("SeqArray")>]
    let seqArray (ms: #seq<Dynamics<_>>) = Dynamics (fun ps ->
        [| for m in ms -> m <|- ps |])
        
    [<CompiledName ("Init")>]
    let init f = Dynamics (fun ps -> 
        if ps.Iteration = 0 && ps.Phase = 0 then
            f <|- ps
        else
            f <|- { ps with Time = (Specs.time ps.Specs 0 0);
                            Iteration = 0; Phase = 0 })
    
    [<CompiledName ("Discrete")>]
    let discrete f = Dynamics (fun ps -> 
    
        if ps.Phase = 0 then
            f <|- ps
        elif ps.Phase > 0 then
            f <|- { ps with Time = (Specs.time ps.Specs ps.Iteration 0); Phase = 0 }
        else
        
            let s = ps.Specs
            let t = ps.Time
            let n = ps.Iteration
        
            let t' = s.StartTime + float (n + 1) * s.DT
            let n' = if Specs.neighborhood s t t' then n + 1 else n
            
            f <|- { ps with Time = (Specs.time s n' 0); Iteration = n'; Phase = 0 })
    
    [<CompiledName ("Interpolate")>]
    let interpolate f = Dynamics (fun ps ->
    
        if ps.Phase >= 0 then
            f <|- ps
            
        else
        
            let s = ps.Specs
            let t = ps.Time
        
            let x = (t - s.StartTime) / s.DT
            
            let n1 = int (floor x)
            let n2 = int (ceil x)
            
            let n1 = max 0 n1
            let n2 = min n2 (s.Iterations - 1)
            
            let t1 = Specs.time s n1 0
            let t2 = Specs.time s n2 0
            
            let y1 = f <|- { ps with Time = t1; Iteration = n1; Phase = 0 }
            let y2 = f <|- { ps with Time = t2; Iteration = n2; Phase = 0 }
            
            if t = t1 then y1
            elif t = t2 then y2
            else y1 + (y2 - y1) * (t - t1) / (t2 - t1))
    
    type Memo<'a> = {mutable iter: int; mutable phase: int; data: 'a [][]}
    
    [<CompiledName ("Memo")>]
    let memo (tr: Dynamics<'a> -> Dynamics<'a>) (f: Dynamics<'a>) =

        let dict = ref Map.empty
        let dictlock = new obj ()
        
        let fix cell (ps: Parameters) =
        
            let n  = ps.Iteration
            let ph = ps.Phase
        
            if (cell.iter < n) || (cell.iter = n) && (cell.phase <= ph) then
            
                let phases = Specs.phases ps.Specs
            
                while (cell.iter < n) || (cell.iter = n) && (cell.phase <= ph) do
                
                    let n'  = cell.iter
                    let ph' = cell.phase
                    let t'  = Specs.time ps.Specs n' ph'
                    
                    let ps' = { ps with Time = t'; Iteration = n'; Phase = ph' }
                    
                    cell.data.[ph'].[n'] <- f <|- ps'
                    
                    if n' <> cell.iter then failwith "memo: recurrent loop"
                    if ph' <> cell.phase then failwith "memo recurrent loop"
                    
                    cell.phase <- cell.phase + 1
                    
                    if (cell.phase >= phases) then
                        cell.phase <- 0
                        cell.iter <- cell.iter + 1
                
            cell.data.[ph].[n]
        
        let create (ps: Parameters) =
        
            let r = ps.Run
            let s = ps.Specs
            let k = r.Id
        
            let iterations = Specs.iterations s
            let phases = Specs.phases s
        
            let data = [| for i = 0 to phases - 1 do yield Array.zeroCreate iterations |]
            let cell = {iter = 0; phase = 0; data = data}

            lock dictlock (fun () -> dict := Map.add k cell !dict)
            r.Disposed.Add (fun () -> lock dictlock (fun () -> dict := Map.remove k !dict))
            cell
                
        in Dynamics (fun ps ->
        
            let r = ps.Run
            let k = r.Id
        
            if Map.containsKey k !dict then
                fix (Map.find k !dict) ps
            elif r.IsDisposed then
                failwith "memo: unexpected call whereas the simulation is complete"
            else
                fix (create ps) ps)
                
        |> tr
    
    type Memo0<'a> = {mutable iter: int; data: 'a []}
    
    [<CompiledName ("Memo0")>]
    let memo0 (tr: Dynamics<'a> -> Dynamics<'a>) (f: Dynamics<'a>) =

        let dict = ref Map.empty
        let dictlock = new obj ()
        
        let fix cell (ps: Parameters) =
        
            let n = ps.Iteration
        
            while cell.iter <= n do
            
                let n'  = cell.iter
                let t'  = Specs.time ps.Specs n' 0
                
                let ps' = { ps with Time = t'; Iteration = n'; Phase = 0 }
                
                cell.data.[n'] <- f <|- ps'
                if n' <> cell.iter then failwith "memo0: recurrent loop"
                cell.iter <- cell.iter + 1
                
            cell.data.[n]
        
        let create (ps: Parameters) =
        
            let r = ps.Run
            let s = ps.Specs
            let k = r.Id
        
            let iterations = Specs.iterations s
        
            let data = Array.zeroCreate iterations
            let cell = {iter = 0; data = data}

            lock dictlock (fun () -> dict := Map.add k cell !dict)
            r.Disposed.Add (fun () -> lock dictlock (fun () -> dict := Map.remove k !dict))
            cell
                
        in Dynamics (fun ps ->
        
            let r = ps.Run
            let k = r.Id
        
            if Map.containsKey k !dict then
                fix (Map.find k !dict) ps
            elif r.IsDisposed then
                failwith "memo0: unexpected call whereas the simulation is complete"
            else
                fix (create ps) ps)
                
        |> tr
    
    type MemoGenerator<'a, 'b> = {mutable iter: int; g: 'a; data: 'b []}
    
    [<CompiledName ("MemoGenerator")>]
    let memoGenerator (f: Specs -> 'a) (m: Dynamics<'a -> 'b>) =

        let dict = ref Map.empty
        let dictlock = new obj ()
        
        let fix cell (ps: Parameters) =
        
            let n = ps.Iteration
        
            while cell.iter <= n do
            
                let n'  = cell.iter
                let t'  = Specs.time ps.Specs n' 0
                
                let ps' = { ps with Time = t'; Iteration = n'; Phase = 0 }
                
                cell.data.[n'] <- (m <|- ps') cell.g
                if n' <> cell.iter then failwith "memoGenerator: recurrent loop"
                cell.iter <- cell.iter + 1
                
            cell.data.[n]
        
        let create (ps: Parameters) =
        
            let r = ps.Run
            let s = ps.Specs
            let k = r.Id
        
            let iterations = Specs.iterations s
        
            let data = Array.zeroCreate iterations
            let cell = {iter = 0; g = f s; data = data}

            lock dictlock (fun () -> dict := Map.add k cell !dict)
            r.Disposed.Add (fun () -> lock dictlock (fun () -> dict := Map.remove k !dict))
            cell
                
        in Dynamics (fun ps ->
        
            let r = ps.Run
            let k = r.Id
        
            if Map.containsKey k !dict then
                fix (Map.find k !dict) ps
            elif r.IsDisposed then
                failwith "memoGenerator: unexpected call whereas the simulation is complete"
            else
                fix (create ps) ps)
                
        |> discrete
   
    [<CompiledName ("Once")>]
    let once (m: Dynamics<unit>) =

        let dict = ref Map.empty
        let dictlock = new obj ()
        
        in Dynamics (fun ps ->
        
            let r = ps.Run
            let k = r.Id
        
            if not (Map.containsKey k !dict) then
            
                lock dictlock (fun () ->
                    dict := Map.add k () !dict)
                r.Disposed.Add (fun () ->
                    lock dictlock (fun () ->
                        dict := Map.remove k !dict))
                        
                m <|- ps)
   
type Dynamics<'a> with

    static member (~+) (a: Dynamics<float>) = a

    static member (~-) (a: Dynamics<float>) = 
        Dynamics (fun ps -> - (a <|- ps))

    static member (+) (a: Dynamics<float>, b: Dynamics<float>) = 
        Dynamics (fun ps -> (a <|- ps) + (b <|- ps))
        
    static member (+) (a: Dynamics<float>, b: float) =
        Dynamics (fun ps -> (a <|- ps) + b)
        
    static member (+) (a: float, b: Dynamics<float>) =
        Dynamics (fun ps -> a + (b <|- ps))
    
    static member (-) (a: Dynamics<float>, b: Dynamics<float>) =
        Dynamics (fun ps -> (a <|- ps) - (b <|- ps))
        
    static member (-) (a: Dynamics<float>, b: float) =
        Dynamics (fun ps -> (a <|- ps) - b)
        
    static member (-) (a: float, b: Dynamics<float>) =
        Dynamics (fun ps -> a - (b <|- ps))
    
    static member (*) (a: Dynamics<float>, b: Dynamics<float>) =
        Dynamics (fun ps -> (a <|- ps) * (b <|- ps))
        
    static member (*) (a: Dynamics<float>, b: float) =
        Dynamics (fun ps -> (a <|- ps) * b)
        
    static member (*) (a: float, b: Dynamics<float>) =
        Dynamics (fun ps -> a * (b <|- ps))
    
    static member (/) (a: Dynamics<float>, b: Dynamics<float>) =
        Dynamics (fun ps -> (a <|- ps) / (b <|- ps))
        
    static member (/) (a: Dynamics<float>, b: float) =
        Dynamics (fun ps -> (a <|- ps) / b)
        
    static member (/) (a: float, b: Dynamics<float>) =
        Dynamics (fun ps -> a / (b <|- ps))
        
    static member Abs (a: Dynamics<float>) = 
        Dynamics (fun ps -> abs (a <|- ps))
    
    static member Acos (a: Dynamics<float>) = 
        Dynamics (fun ps -> acos (a <|- ps))
    
    static member Asin (a: Dynamics<float>) = 
        Dynamics (fun ps -> asin (a <|- ps))
    
    static member Atan (a: Dynamics<float>) = 
        Dynamics (fun ps -> atan (a <|- ps))
    
    static member Atan2 (a: Dynamics<float>, b: Dynamics<float>) = 
        Dynamics (fun ps -> atan2 (a <|- ps) (b <|- ps))

    static member Ceiling (a: Dynamics<float>) = 
        Dynamics (fun ps -> ceil (a <|- ps))
    
    static member Exp (a: Dynamics<float>) = 
        Dynamics (fun ps -> exp (a <|- ps))
    
    static member Floor (a: Dynamics<float>) = 
        Dynamics (fun ps -> floor (a <|- ps))
    
    static member Truncate (a: Dynamics<float>) = 
        Dynamics (fun ps -> truncate (a <|- ps))
    
    static member Round (a: Dynamics<float>) = 
        Dynamics (fun ps -> round (a <|- ps))
    
    static member Log (a: Dynamics<float>) = 
        Dynamics (fun ps -> log (a <|- ps))
    
    static member Log10 (a: Dynamics<float>) = 
        Dynamics (fun ps -> log10 (a <|- ps))
    
    static member Sqrt (a: Dynamics<float>) = 
        Dynamics (fun ps -> sqrt (a <|- ps))
    
    static member Cos (a: Dynamics<float>) = 
        Dynamics (fun ps -> cos (a <|- ps))
    
    static member Cosh (a: Dynamics<float>) = 
        Dynamics (fun ps -> cosh (a <|- ps))
    
    static member Sin (a: Dynamics<float>) = 
        Dynamics (fun ps -> sin (a <|- ps))
    
    static member Sinh (a: Dynamics<float>) = 
        Dynamics (fun ps -> sinh (a <|- ps))
    
    static member Tan (a: Dynamics<float>) = 
        Dynamics (fun ps -> tan (a <|- ps))
    
    static member Tanh (a: Dynamics<float>) = 
        Dynamics (fun ps -> tanh (a <|- ps))
    
    static member Pow (a: Dynamics<float>, b: Dynamics<float>) = 
        Dynamics (fun ps -> (a <|- ps) ** (b <|- ps))
    
    static member Pow (a: Dynamics<float>, b: float) = 
        Dynamics (fun ps -> (a <|- ps) ** b)
