
// DynamicsCont.fs - Aivika

// Copyright (c) 2009, 2010  David Sorokin. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials
//     provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
// THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.

//
// Author: David Sorokin <david.sorokin@gmail.com>
//

namespace Maritegra.Aivika

open System

type internal DynamicsContFun<'a> = Dynamics<('a -> unit) * (exn -> unit)> -> Dynamics<unit>

[<Sealed>]
type DynamicsCont<'a> (f: DynamicsContFun<'a>) =
    member internal x.Fun = f
    
[<AutoOpen>]
module internal DynamicsContOp =
    let inline (<|--) (m: DynamicsCont<'a>) x = m.Fun x

module DynamicsContBuilderImpl =

    let protectDC f x cont econt =

        let mutable a = Unchecked.defaultof<_>
        let mutable e = null
        
        try
            a <- f x
        with
        | exn -> 
            e <- exn
        
        match e with
        | null -> cont a
        | exn -> econt exn

    let inline returnDC a = DynamicsCont (fun c -> 
        Dynamics (fun ps -> 
        
            let (cont, econt) = c <|- ps
            in cont a))

    let inline bindDC m k = DynamicsCont (fun c -> 
        m <|-- (Dynamics (fun ps ->
        
            let (cont, econt) = c <|- ps
            let cont a = protectDC k a (fun m -> (m <|-- c) <|- ps) econt
            in (cont, econt))))

    // letDC a k = bindDC (returnDC a) k
    let inline letDC a k = DynamicsCont (fun c ->
        Dynamics (fun ps ->
        
            let (cont, econt) = c <|- ps
            in protectDC k a (fun m -> (m <|-- c) <|- ps) econt))

    //  delayDC k = bindDC (returnDC ()) k
    let inline delayDC k = letDC () k

    //  callDC k a = bindDC (returnDC a) k        
    let inline callDC k a = letDC a k

    let zeroDC = returnDC ()
    
    let inline tryFinallyDC m f = DynamicsCont (fun c ->
        m <|-- (Dynamics (fun ps ->
            
            let (cont, econt) = c <|- ps
            let cont' a = protectDC f () (fun () -> cont a) econt
            let econt' e = protectDC f () (fun () -> econt e) econt
            in (cont', econt'))))       
    
    let inline tryWithDC m k = DynamicsCont (fun c ->
        m <|-- (Dynamics (fun ps ->
        
            let (cont, econt) = c <|- ps
            let econt' e = ((callDC k e) <|-- c) <|- ps
            (cont, econt'))))

    let inline usingDC (r: 'a :> IDisposable) k =  
        tryFinallyDC (callDC k r) (fun () -> r.Dispose())

    let rec whileDC p m =
        if p () then 
            bindDC m (fun () -> whileDC p m) 
        else 
            zeroDC
            
    let rec forDC (es: seq<_>) m =
        usingDC (es.GetEnumerator()) (fun ie ->
            whileDC
                (fun () -> ie.MoveNext())
                (delayDC (fun () -> m ie.Current)))

    let inline combineDC m1 m2 = 
        bindDC m1 (fun () -> m2)

open DynamicsContBuilderImpl

[<Sealed>]
type DynamicsContBuilder() =

    member d.Return (a) = returnDC a
    member d.ReturnFrom (m: DynamicsCont<_>) = m
    member d.Bind (m, k) = bindDC m k
    member d.Delay (k) = delayDC k
    member d.Zero () = zeroDC
    member d.Combine (m1, m2) = combineDC m1 m2
    member d.For (es, k) = forDC es k
    member d.While (p, m) = whileDC p m
    member d.Using (a, k) = usingDC a k
    member d.TryFinally (m, f) = tryFinallyDC m f
    member d.TryWith (m, k) = tryWithDC m k

[<AutoOpen>]
module DynamicsContWorkflow =
    let dynamicscont = new DynamicsContBuilder()

[<RequireQualifiedAccess>]
[<CompilationRepresentation (CompilationRepresentationFlags.ModuleSuffix)>]
module DynamicsCont =

    [<CompiledName ("Lift")>]
    let lift d = DynamicsCont (fun c ->
        Dynamics (fun ps ->
            
            let (cont, econt) = c <|- ps
            in protectDC (fun () -> d <|- ps) () (fun a -> cont a) econt))
            
    [<CompiledName ("Run")>]
    let run m f = m <|-- Dynamics (fun ps -> (f, fun e -> raise e))
    
    [<CompiledName ("Eta")>]
    let eta a = dynamicscont {
        return a
    }
    
    [<CompiledName ("Mu")>]
    let mu mm = dynamicscont {
        let! m = mm in return! m
    }
    
    [<CompiledName ("Map")>]
    let map f m = dynamicscont {
        let! a = m 
        return f a
    }
    
    [<CompiledName ("Ap")>]
    let ap mf m = dynamicscont {
        let! f = mf
        let! a = m
        return f a
    }
    
    [<CompiledName ("Lift2")>]
    let lift2 f m1 m2 = dynamicscont {
        let! a1 = m1
        let! a2 = m2
        return f a1 a2
    }
    
    [<CompiledName ("Lift3")>]
    let lift3 f m1 m2 m3 = dynamicscont {
        let! a1 = m1
        let! a2 = m2
        let! a3 = m3
        return f a1 a2 a3
    }
    
    [<CompiledName ("Lift4")>]
    let lift4 f m1 m2 m3 m4 = dynamicscont {
        let! a1 = m1
        let! a2 = m2
        let! a3 = m3
        let! a4 = m4
        return f a1 a2 a3 a4
    }
    
    [<CompiledName ("Lift5")>]
    let lift5 f m1 m2 m3 m4 m5 = dynamicscont {
        let! a1 = m1
        let! a2 = m2
        let! a3 = m3
        let! a4 = m4
        let! a5 = m5
        return f a1 a2 a3 a4 a5
    }
        
    [<CompiledName ("Zip")>]
    let zip m1 m2 = dynamicscont {
        let! a1 = m1
        let! a2 = m2
        return (a1, a2)
    }
        
    [<CompiledName ("Zip3")>]
    let zip3 m1 m2 m3 = dynamicscont {
        let! a1 = m1
        let! a2 = m2
        let! a3 = m3
        return (a1, a2, a3)
    }
