
// PriorityQueue.fs - Aivika

// Copyright (c) 2009, 2010  David Sorokin. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials
//     provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
// THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.

//
// Author: David Sorokin <david.sorokin@gmail.com>
//

namespace Maritegra.Aivika

open System

type internal PriorityQueue<'a> () = 

    let mutable ks: float [] = Array.zeroCreate 11
    let mutable vs: 'a [] = Array.zeroCreate 11
    
    let kref = ref ks
    let vref = ref vs
    
    let mutable size = 0
    
    let increase capacity =
    
        let len = ks.Length
        
        assert (capacity >= 0)
        let capacity' = if len < 64 then (len + 1) <<< 1 else (len >>> 1) * 3
        assert (capacity' >= 0)
        let capacity' = if capacity' < capacity then capacity else capacity'
        
        Array.Resize (kref, capacity')
        Array.Resize (vref, capacity')
        
        ks <- !kref
        vs <- !vref
    
    let rec siftUp i k v = 
        if i = 0 then
            ks.[i] <- k
            vs.[i] <- v
        else
            let n = (i - 1) >>> 1
            if k >= ks.[n] then
                ks.[i] <- k
                vs.[i] <- v
            else
                ks.[i] <- ks.[n]
                vs.[i] <- vs.[n]
                siftUp n k v    // tail call
        
    let rec siftDown i k v =
        if i >= (size >>> 1) then
            ks.[i] <- k
            vs.[i] <- v
        else
            let n  = (i <<< 1) + 1
            let n' = n + 1
            let n  = if n' < size && ks.[n] > ks.[n'] then n' else n
            if k <= ks.[n] then
                ks.[i] <- k
                vs.[i] <- v
            else
                ks.[i] <- ks.[n]
                vs.[i] <- vs.[n]
                siftDown n k v  // tail call

    member x.IsEmpty = (size = 0)

    member x.Enqueue (k, v) =
    
        let i = size
        if i >= ks.Length then 
            increase (i + 1)
        size <- i + 1
        
        siftUp i k v

    member x.Dequeue () =
    
        assert (size > 0)
        size <- size - 1
        
        let k = ks.[size]
        let v = vs.[size]
        
        ks.[size] <- Unchecked.defaultof<_>
        vs.[size] <- Unchecked.defaultof<_>
        
        if size > 0 then
            siftDown 0 k v

    member x.Front =
        assert (size > 0)
        (ks.[0], vs.[0])
        
    member x.FrontKey =
        assert (size > 0)
        ks.[0]
        
    member x.FrontValue =
        assert (size > 0)
        vs.[0]