﻿
(* This model is described in the Russian book stated below.
  
   Книга "Объектно-ориентированное моделирование на C++", 2006. Автор Илья Труб.
   Модель описана в главе 6 "Регулярный входной поток, отсутствие очередей,
   естественный отсчет времени: моделирование больничной палаты". 
   
   По своей сути результаты соответствуют описанным в книге. Суммы основных
   параметров в целом совпадают. Различие идет в деталях того, как пациенты
   выписываются из палаты.
   
   Можно обратить внимание на то, как вычисляется среднее значение
   через интеграл. То есть, модель фактически является гибридной. 
   Используются методы DES и системной динамики вместе. *)

open System

open Maritegra.Aivika
open Maritegra.Aivika.SD

module Hospital =

    let specs = {
    
        StartTime=0.0; StopTime=32014.0; DT=1.0; 
        Method=RungeKutta4; Randomness=SimpleRnd
    }

    let env = Env ()
    
    let queueSize = 25
    
    let mutable patientNum = 0
    let mutable stayNum = 0
    let mutable acceptNum = 0
    
    let mutable reject41Num = 0
    let mutable rejectNum = 0
    
    let mutable dischargeNum = 0
    let mutable cureNum = 0
    
    let mutable loadNum = 0.0
    
    let rnd = new Random ()
    
    let healthvariateI () =
        30.0 + 14.0 * rnd.NextDouble ()
        
    let healthvariate h =
        h + 1.4 * rnd.NextDouble () - 0.2
    
    type Patient (e: Env) =
    
        inherit Process (e)
        
        static let queue = new ResizeArray<Patient> ()
        
        let isQueueFull () = queue.Count >= queueSize
        
        let discharge () =
            let i = queue.Count - 1
            if (i >= 0) && (queue.[i].Health >= 47.0) then
                queue.[i].Discharged <- true
                queue.RemoveAt (i)
        
        let rec sort i =
            if (i > 0) && (queue.[i - 1].Health > queue.[i].Health) then
                let p = queue.[i]
                queue.[i] <- queue.[i - 1]
                queue.[i - 1] <- p
                queue.[i].Index <- i
                queue.[i - 1].Index <- i - 1
                sort (i - 1)
            elif (i < queue.Count - 1) && (queue.[i + 1].Health < queue.[i].Health) then
                let p = queue.[i]
                queue.[i] <- queue.[i + 1]
                queue.[i + 1] <- p
                queue.[i].Index <- i
                queue.[i + 1].Index <- i + 1
                sort (i + 1)
                
        let add (p: Patient) =
            queue.Add (p)
            p.Index <- queue.Count - 1
            sort p.Index
            
        let remove i =
            queue.RemoveAt (i)
            for j = i to queue.Count - 1 do
                queue.[j].Index <- j
        
        let mutable index = -1
        let mutable health = 1.0
        let mutable discharged = false
        
        member private x.Health = health
        member private x.Index with get () = index and set (v) = index <- v
        member private x.Discharged with set (v) = discharged <- v
        
        override x.Activate = dynamicscont {
            
            patientNum <- patientNum + 1
            health <- healthvariateI ()
            
            if (health > 41.0) && isQueueFull () then
                reject41Num <- reject41Num + 1
            else
                
                if isQueueFull () then
                    discharge ()
                
                if isQueueFull () then
                    rejectNum <- rejectNum + 1
                else
            
                    acceptNum <- acceptNum + 1
                
                    add x
                    do! x.Heal ()
        }
        
        member private x.Heal () = dynamicscont {
        
            loadNum <- float queue.Count / float queueSize
        
            do! x.HoldD (dt)
            
            if discharged then
                dischargeNum <- dischargeNum + 1
            else
            
                stayNum <- stayNum + 1
                health <- healthvariate health
                
                assert obj.ReferenceEquals (x, queue.[index])
                sort index
                assert obj.ReferenceEquals (x, queue.[index])
               
                if health > 49.0 then
                
                    remove index
                    cureNum <- cureNum + 1
                    
                else
                    do! x.Heal ()
        }
        
    type Hospital (env: Env) =
    
        inherit Process (env)
        
        override x.Activate = dynamicscont {
        
            while true do
            
                let p1 = Patient (env)
                let p2 = Patient (env)
                
                do! DynamicsCont.lift <| p1.StartD (time)
                do! DynamicsCont.lift <| p2.StartD (time)
                
                // ---
                if patientNum % 1000 = 0 then
                    printfn "Число претендентов: %i" patientNum
                // ---
                
                do! x.HoldD (dt)
        }
        
    let hospital = Hospital (env)
    
    let mean m = integ (lazy m) (eta 0.0) / (stoptime - starttime)
    
    let loadMean = mean (dynamics { return loadNum })
    
    let starter = hospital.StartD (starttime)
    
    let system = dynamics {
    
        do! starter
        do! env.Run
        
        let! loadVal = loadMean
        
        return (patientNum, cureNum, reject41Num, rejectNum, 
            dischargeNum, acceptNum, loadVal, (float) stayNum / (float) acceptNum)
    }
    
    [<EntryPoint>]
    let main args =
    
        let result = Dynamics.runLast system specs
        let (x1, x2, x3, x4, x5, x6, x7, x8) = result
        
        printfn "Общее число претендентов: %i" x1
        printfn "Завершили лечение: %i" x2
        printfn "Не приняты на лечение (> 41 балла): %i" x3
        printfn "Не приняты на лечение: %i" x4
        printfn "Выписаны досрочно: %i" x5
        printfn "Всего принято на лечение: %i" x6
        printfn "Средняя загрузка палаты: %f" x7
        printfn "Среднее время пребывания в палате: %f" x8
    
        0
