﻿
(* This is a rewritten version of model MachRep1 based on
   the activity-oriented paradigm of DES. *)

(* It corresponds to model MachRep1 described in document 
   Introduction to Discrete-Event Simulation and the SimPy Language
   [http://heather.cs.ucdavis.edu/~matloff/156/PLN/DESimIntro.pdf]. 
   SimPy is available on [http://simpy.sourceforge.net/].
   
   The model description is as follows.

   Two machines, which sometimes break down.
   Up time is exponentially distributed with mean 1.0, and repair time is
   exponentially distributed with mean 0.5. There are two repairpersons,
   so the two machines can be repaired simultaneously if they are down
   at the same time.

   Output is long-run proportion of up time. Should get value of about
   0.66. *)

open System

open Maritegra.Aivika
open Maritegra.Aivika.SD

let specs = {

    StartTime=0.0; StopTime=10000.0; DT=0.05; 
    Method=RungeKutta4; Randomness=SimpleRnd
}

let upRate = 1.0 / 1.0          // reciprocal of mean up time
let repairRate = 1.0 / 0.5      // reciprocal of mean repair time

let mutable totalUpTime = 0.0   // total up time for all machines

let expovariate = 
    let rnd = new Random ()
    in fun lambda ->
        - log (rnd.NextDouble ()) / lambda 
        
let machine id = 

    let upTimeNum = ref -1
    let repairTimeNum = ref -1
    let startUpTime = ref 0.0
    
    dynamics {
    
        if !upTimeNum > 0 then
            decr upTimeNum
            
        if !upTimeNum = 0 then
            upTimeNum := -1
                
            // the machine is broken
            let! finishUpTime = time
            let! dt' = dt
            totalUpTime <- totalUpTime 
                + (finishUpTime - !startUpTime)
            repairTimeNum := int <| 
                (expovariate repairRate) / dt'
            
        elif !upTimeNum < 0 then
        
            if !repairTimeNum > 0 then
                decr repairTimeNum
                
            if !repairTimeNum = 0 then
                repairTimeNum := -1
                
                // the machine is repaired
                let! t' = time
                let! dt' = dt
                startUpTime := t'
                upTimeNum := int <| (expovariate upRate) / dt'
                
            elif !repairTimeNum < 0 then
                
                // initialization
                let! t' = time
                let! dt' = dt
                startUpTime := t'
                upTimeNum := int <| (expovariate upRate) / dt'
                
    } |> Dynamics.memo0 discrete
    
let machine1 = machine 1
let machine2 = machine 2

let system = dynamics {

    do! machine1    // process machine 1
    do! machine2    // process machine 2
    return! (totalUpTime / (2.0 * stoptime))
}

[<EntryPoint>]
let main args =

    let result = Dynamics.runLast system specs
    
    printfn "Long-run proportion of up time = %f (must be about 0.66)" result
    
    0
