﻿
(* This model is described in the Russian book stated below.
  
   Книга "Объектно-ориентированное моделирование на C++", 2006. Автор Илья Труб.
   Модель описана в главе 8 "Последовательное обслуживание с возвращениями:
   производственная линия с пунктами технического контроля и настройки". 
   
   Можно обратить внимание на то, как вычисляется среднее значение
   через интеграл. То есть, модель фактически является гибридной. 
   Используются методы DES и системной динамики вместе. *)

#nowarn "40"

open System
open System.Collections.Generic

open Maritegra.Aivika
open Maritegra.Aivika.SD

module TvControl =

    let specs = {
    
        StartTime=0.0; StopTime=480.0; DT=1.00; 
        Method=Euler; Randomness=SimpleRnd
    }
    
    let inline lift m = DynamicsCont.lift m
    
    let mean m = integ (lazy m) (eta 0.0) / (time - starttime)
    
    let entryMinTime = 3.5
    let entryMaxTime = 7.5
    
    let controlMinTime = 6.0
    let controlMaxTime = 12.0
    
    let controlRate = 0.85
    
    let tuningMinTime = 20.0
    let tuningMaxTime = 40.0
    
    let rnd = Random ()
    
    let entryTime () =
        entryMinTime + (entryMaxTime - entryMinTime) * rnd.NextDouble ()
    
    let controlTime () =
        controlMinTime + (controlMaxTime - controlMinTime) * rnd.NextDouble ()
        
    let controlTest () =
        rnd.NextDouble () <= controlRate
        
    let tuningTime () =
        tuningMinTime + (tuningMaxTime - tuningMinTime) * rnd.NextDouble ()
    
    let controlQueue = ResizeArray<_> ()
    let tuningQueue = ResizeArray<_> ()
    
    let env = Env ()
    
    let totalNum = Var (env, 0)
    
    let processSum = Var (env, 0.0)
    let processNum = Var (env, 0)
    let processTime = processSum.Value / Dynamics.map float processNum.Value
    
    let controlQueueSize = Var (env, 0.0)
    let controlQueueSizeMean = mean controlQueueSize.Value
    
    let tuningQueueSize = Var (env, 0.0)
    let tuningQueueSizeMean = mean tuningQueueSize.Value
    
    let controlWkLoad = Var (env, 0.0)
    let controlWkLoadMean = mean controlWkLoad.Value
    
    let tuningWkLoad = Var (env, 0.0)
    let tuningWkLoadMean = mean tuningWkLoad.Value
    
    type ControlWorkplace (e, tuningWk: Lazy<TuningWorkplace>) =
        
        inherit Process (e)
        
        member x.Pass (t) = dynamicscont {
            let! t2 = time |> lift
            do! processSum |> Var.incValue (t2 - t) |> lift
            do! processNum |> Var.incValue 1 |> lift
        }
        
        override x.Activate = dynamicscont {
            while true do
                if controlQueue.Count = 0 then
                    do! x.Passivate
                else
                    // -- get --
                    let t = controlQueue.[0]
                    controlQueue.RemoveAt (0)
                    do! controlQueueSize.SetValue (float controlQueue.Count) |> lift
                    // -- load --
                    do! controlWkLoad |> Var.incValue 0.5 |> lift
                    do! x.Hold (controlTime ())
                    do! controlWkLoad |> Var.decValue 0.5 |> lift
                    // -- test --
                    if controlTest () then
                        do! x.Pass (t)
                    else
                        do! tuningWk.Value.Load (t)
        }
        
    and TuningWorkplace (e, line: Lazy<Line>) =
        
        inherit Process (e)
        
        member x.Load (t) = dynamicscont {
            tuningQueue.Add (t)
            do! tuningQueueSize |> Var.incValue 1.0 |> lift
            if tuningQueue.Count = 1 then
                do! x.Reactivate
        }
        
        override x.Activate = dynamicscont {
            while true do
                if tuningQueue.Count = 0 then
                    do! x.Passivate
                else
                    // -- get --
                    let t = tuningQueue.[0]
                    tuningQueue.RemoveAt (0)
                    do! tuningQueueSize.SetValue (float tuningQueue.Count) |> lift
                    // -- load --
                    do! tuningWkLoad.SetValue (1.0) |> lift
                    do! x.Hold (tuningTime ())
                    do! tuningWkLoad.SetValue (0.0) |> lift
                    // -- again --
                    do! line.Value.Load (t)
        }
        
    and Line (e, controlWk1: Lazy<ControlWorkplace>, controlWk2: Lazy<ControlWorkplace>, tuningWk: Lazy<TuningWorkplace>) =
    
        inherit Process (e)
        
        member x.Load (t) = dynamicscont {
            controlQueue.Add (t)
            do! controlQueueSize |> Var.incValue 1.0 |> lift
            if controlQueue.Count = 1 then
                do! controlWk1.Value.Reactivate
                do! controlWk2.Value.Reactivate
        }
        
        override x.Activate = dynamicscont {
            while true do
                // -- entry --
                do! x.Hold (entryTime ())
                do! totalNum |> Var.incValue 1 |> lift
                // -- load --
                let! t = time |> lift
                do! x.Load (t)
        }
        
    let rec controlWk1 = ControlWorkplace (env, lazy tuningWk)
    and controlWk2 = ControlWorkplace (env, lazy tuningWk)
    and tuningWk = TuningWorkplace (env, lazy line)
    and line = Line (env, lazy controlWk1, lazy controlWk2, lazy tuningWk)
    
    let starter = 
        dynamics {
    
            do! controlWk1.StartD (starttime)
            do! controlWk2.StartD (starttime)
            do! tuningWk.StartD (starttime)
            do! line.StartD (starttime)
            
        } |> Dynamics.once
    
    let system = dynamics {
    
        do! starter
        do! env.Run
        
        let! x1 = totalNum.Value
        let! x2 = processNum.Value
        let! x3 = controlWkLoadMean
        let! x4 = tuningWkLoadMean
        let! x5 = processTime
        let! x6 = controlQueueSizeMean
        let! x7 = tuningQueueSizeMean
        
        return (x1, x2, x3, x4, x5, x6, x7)
    }

    [<EntryPoint>]    
    let main args =
    
        let (x1, x2, x3, x4, x5, x6, x7) = Dynamics.runLast system specs
    
        printfn "total number: %i" x1
        printfn "processed number: %i" x2
        printfn "control load rate: %f" x3
        printfn "tuning load rate: %f" x4
        printfn "processing time: %f" x5
        printfn "control queue size: %f" x6
        printfn "tuning queue size: %f" x7
        
        0
    
